import schemdraw
from typing import Callable, Type, TypeVar 
from . import Elements as elm
from ..Network.network import Branch
from ..Network import elements as network_elmements

SchemdrawElement = TypeVar('SchemdrawElement', bound=schemdraw.elements.Element)
SchemdrawElementTranslator = Callable[[SchemdrawElement, Callable[[schemdraw.util.Point], str]], tuple[Branch, str]]

def round_node(node: schemdraw.util.Point) -> schemdraw.util.Point:
    def local_round(x):
        return round(x, ndigits=2)
    return schemdraw.util.Point((local_round(node.x), local_round(node.y)))

def get_nodes(element: schemdraw.elements.Element, n1_label='start', n2_label='end') -> tuple[schemdraw.util.Point, schemdraw.util.Point]:
    return round_node(element.absanchors[n1_label]), round_node(element.absanchors[n2_label])

def real_current_source_translator(element: elm.RealCurrentSource, node_mapper: Callable[[schemdraw.util.Point], str]) -> tuple[Branch, str]:
    n1, n2 = get_nodes(element)
    return Branch(node_mapper(n1), node_mapper(n2), network_elmements.real_current_source(element.I, element.R)), element.name

def resistor_translator(element: elm.Resistor, node_mapper: Callable[[schemdraw.util.Point], str]) -> tuple[Branch, str]:
    n1, n2 = get_nodes(element)
    return Branch(node_mapper(n1), node_mapper(n2), network_elmements.resistor(element.R)), element.name

def current_source_translator(element: elm.CurrentSource, node_mapper: Callable[[schemdraw.util.Point], str]) -> tuple[Branch, str]:
    n1, n2 = get_nodes(element)
    return Branch(node_mapper(n1), node_mapper(n2), network_elmements.current_source(element.I)), element.name

def real_voltage_source_translator(element: elm.RealVoltageSource, node_mapper: Callable[[schemdraw.util.Point], str]) -> tuple[Branch, str]:
    n1, n2 = get_nodes(element)
    return Branch(node_mapper(n1), node_mapper(n2), network_elmements.real_voltage_source(element.V, element.R)), element.name

def voltage_source_translator(element: elm.VoltageSource, node_mapper: Callable[[schemdraw.util.Point], str]) -> tuple[Branch, str]:
    n1, n2 = get_nodes(element)
    return Branch(node_mapper(n2), node_mapper(n1), network_elmements.voltage_source(element.V)), element.name

def new_real_current_source_translator(element: elm.RealCurrentSource, node_mapper: Callable[[schemdraw.util.Point], str]) -> tuple[Branch, str]:
    n1, n2 = get_nodes(element)

    return Branch(node_mapper(n1), node_mapper(n2), network_elmements.real_current_source(element.I, element.R)), element.name

element_translator : dict[Type[schemdraw.elements.Element], SchemdrawElementTranslator] = {
    elm.Resistor : resistor_translator,
    elm.CurrentSource: current_source_translator,
    elm.VoltageSource: voltage_source_translator,
    elm.RealCurrentSource: new_real_current_source_translator,
    elm.RealVoltageSource: real_voltage_source_translator
}

def translator_available(element: schemdraw.elements.Element) -> bool:
    return type(element) in element_translator.keys()


from ..Utils import scientific_float

blue = '#02468F'
red = '#D20000'

def print_current(I: complex, precision: int = 3) -> str:
    real_part = scientific_float(I.real, precision=precision, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'})
    if I.imag/precision < 1:
        return real_part
    else:
        imag_part = scientific_float(I.imag, precision=precision, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'})
        return f'{real_part} + j{imag_part}'

def print_voltage(V: complex, precision: int = 3) -> str:
    real_part = scientific_float(V.real, precision=precision, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'})
    if V.imag/precision < 1:
        return real_part
    else:
        imag_part = scientific_float(V.imag, precision=precision, use_exp_prefix=True, exp_prefixes={-6: 'u', -3: 'm', 3: 'k'})
        return f'{real_part} + j{imag_part}'


import numpy as np

exp_prefixes = {
    -12 : 'p',
    -9 : 'n',
    -6 : 'u',
    -3 : 'm',
    -1 : 'c',
    3 : 'k',
    6 : 'M',
    9 : 'G',
    12 : 'T'
}

def scientific_float(value: float, precision: int = 3, use_exp_prefix: bool = False, exp_prefixes: dict[int, str]=exp_prefixes) -> str:
    def _floor3(value: float) -> int:
        return 3*np.floor(value/3)
    def _left_comma_valid_digits(value: float) -> int:
        return int(np.floor(np.log10(np.abs(value))))+1
    exp_extension = ''
    exp_prefix = ''
    scaling_exp = int(_floor3(np.log10(np.abs(value))))
    scaling = 10**scaling_exp
    value /= scaling
    decimal_places = max(0, precision - _left_comma_valid_digits(value))
    if use_exp_prefix:
        if scaling_exp in exp_prefixes.keys():
            exp_prefix = exp_prefixes[scaling_exp]
            scaling_exp -= scaling_exp
        else:
            if scaling_exp > max(exp_prefixes.keys()):
                exp_prefix = exp_prefixes[max(exp_prefixes.keys())]
                scaling_exp -= max(exp_prefixes.keys())
            if scaling_exp < min(exp_prefixes.keys()):
                exp_prefix = exp_prefixes[min(exp_prefixes.keys())]
                scaling_exp -= min(exp_prefixes.keys())
    if scaling_exp != 0:
        exp_extension = f'e{scaling_exp}'
    return f'{value:.{decimal_places}f}{exp_extension}{exp_prefix}'


import schemdraw
from .Display import red, blue, print_voltage, print_current
from typing import Any

class Schematic(schemdraw.Drawing):
    def save_copy(self, fname: str, **kwargs) -> None:
        import copy
        cpy = copy.deepcopy(self)
        cpy.save(fname, **kwargs)

class VoltageSource(schemdraw.elements.sources.SourceV):
    def __init__(self, V: float, name: str, *args, reverse=False, precision=3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        if reverse:
            self._V = -V
        else:
            self._V = V
        self._name = name
        self.label(f'{self._name}={print_voltage(V, precision=precision)}V', rotate=True)
        self.segments = DrawVoltageSource()

        a, b = (1.5, 0.7), (-0.5, 0.7)
        self.segments.append(schemdraw.Segment((a, b), arrow='->', arrowwidth=.3, arrowlength=.4, color=blue))

    @property
    def name(self) -> str:
        return self._name

    @property
    def V(self) -> float:
        return self._V

    def values(self) -> dict[str, float]:
        return {'U' : self.V}

class CurrentSource(schemdraw.elements.sources.SourceI):
    def __init__(self, I: float, name: str, *args, reverse=False, precision=3, **kwargs):
        super().__init__(*args, reverse=reverse, **kwargs)
        if reverse:
            self._I = -I
        else:
            self._I = I
        self._name = name
        self.label(f'{self._name}={print_current(I, precision=precision)}A', rotate=True)
        self.segments = DrawCurrentSource()

        a, b = (1.2, -0.3), (1.8, -0.3)
        self.segments.append(schemdraw.Segment((a, b), arrow='->', arrowwidth=.3, arrowlength=.4, color=red))

    @property
    def name(self) -> str:
        return self._name

    @property
    def I(self) -> float:
        return self._I

    def values(self) -> dict[str, float]:
        return {'I' : self.I}

def DrawSource() -> list[schemdraw.segments.SegmentType]:
    return [
        schemdraw.segments.Segment([(0, 0), (0, 0), schemdraw.elements.elements.gap, (1, 0), (1, 0)]),
        schemdraw.segments.SegmentCircle((0.5, 0), 0.5,)
    ]

def DrawCurrentSource() -> list[schemdraw.segments.SegmentType]:
    return DrawSource() + [
        schemdraw.segments.Segment([(0.5, -0.5), (0.5, 0.5)])
    ]

def DrawRealCurrentSource() -> list[schemdraw.segments.SegmentType]:
    return DrawCurrentSource() + [
        schemdraw.segments.SegmentCircle((-0.5, 0), 0.07, fill='black'),
        schemdraw.segments.SegmentCircle((1.5, 0), 0.07, fill='black'),
        schemdraw.segments.Segment([(-0.5, 0), (-0.5, 1)]),
        schemdraw.segments.Segment([(1.5, 0), (1.5, 1)]),
        schemdraw.segments.Segment([(-0.5, 1), (0, 1)]),
        schemdraw.segments.Segment([(1, 1), (1.5, 1)]),
        schemdraw.segments.Segment([(0, 0.8), (1, 0.8), (1, 1.2), (0, 1.2), (0, 0.8)])
    ]

def DrawVoltageSource() -> list[schemdraw.segments.SegmentType]:
    return DrawSource() + [
        schemdraw.segments.Segment([(0, 0), (1, 0)])
    ]

class Resistor(schemdraw.elements.twoterm.ResistorIEC):
    def __init__(self, R: float, name: str, *args, show_name: bool = True, show_value: bool = True, label_offset: float = 0.2, **kwargs):
        super().__init__(*args, **kwargs)
        self._R = R
        self._name = name
        label = ''
        label += f'{self._name}' if show_name else ''
        label += '=' if  show_name and show_value else ''
        label += f'{self.R}$\\Omega$' if show_value else ''
        self.label(label, rotate=True, ofst=label_offset)

    @property
    def name(self) -> str:
        return self._name

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    def values(self) -> dict[str, float]:
        return {'R' : self._R}

    def _place_label(self, *args, **kwargs):
        delta = self.end-self.start
        if abs(delta[1]) > abs(delta[0]): # portrait placing of resistor
            if delta[1] < 0:
                kwargs.update({'rotation': 90})

        super()._place_label(*args, **kwargs)

class RealCurrentSource(schemdraw.elements.compound.ElementCompound):
    def __init__(self, current_source: CurrentSource, resistor: Resistor, *args, d:str = 'up', **kwargs):
        super().__init__(*args, **kwargs)
        if d == 'left':
            self.add(current_source.left())
            self.add(Line('up'))
            self.add(resistor.right())
            self.add(Line('down'))
        elif d == 'down':
            self.add(current_source.down())
            self.add(Line('right'))
            self.add(resistor.up())
            self.add(Line('left'))
        elif d == 'right':
            self.add(current_source.right())
            self.add(Line('up'))
            self.add(resistor.left())
            self.add(Line('down'))
        else:
            self.add(current_source)
            self.add(Line('right'))
            self.add(resistor.down())
            self.add(Line('left'))
        self.anchors['start'] = current_source.start
        self.anchors['end'] = current_source.end
        self.drop(current_source.end)
        self._name = current_source.name
        self._I = current_source.I
        self._R = resistor.R

    @property
    def name(self) -> str:
        return self._name

    @property
    def I(self) -> float:
        return self._I

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    def values(self) -> dict[str, float]:
        return {'I' : self.I, 'R' : self.R}

class RealVoltageSource(schemdraw.elements.compound.ElementCompound):
    def __init__(self, voltage_source: VoltageSource, resistor: Resistor, *args, reverse=False, **kwargs):
        super().__init__(*args, d='right', **kwargs)
        if reverse:
            self.add(resistor)
            self.add(voltage_source)
            start = resistor.start
            end = voltage_source.end
        else:
            self.add(voltage_source)
            self.add(resistor)
            start = voltage_source.start
            end = resistor.end
        self.anchors['start'] = start
        self.anchors['end'] = end
        self.anchors['center'] = (resistor.end-resistor.start)/2 + resistor.start
        self.drop(end)
        self._name = voltage_source.name
        self._V = voltage_source.V
        self._R = resistor.R
        self.resistor_length = resistor._userparams.get('l', resistor._userparams.get('unit', 3))
        self.resistor_d = resistor._userparams.get('d', 'right')

    @property
    def name(self) -> str:
        return self._name

    @property
    def V(self) -> float:
        return self._V

    @property
    def R(self) -> float:
        return self._R

    @property
    def G(self) -> float:
        return 1/self._R

    def values(self) -> dict[str, float]:
        return {'I' : self.I, 'R' : self.R}

class Line(schemdraw.elements.lines.Line):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def name(self) -> str:
        return ''

class Node(schemdraw.elements.Element):
    def __init__(self, id: str = '', *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.node_id = id
        self.params['theta'] = 0
        self.params['drop'] = (0, 0)
        self.anchors['start'] = (0, 0)
        self.anchors['center'] = (0, 0)
        self.anchors['end'] = (0, 0)

    @property
    def name(self) -> str:
        return f'Node {self.node_id}'

class LabelNode(Node):
    def __init__(self, id : str = '', id_loc : str | dict[str, Any] = '', *args, show=True, **kwargs):
        super().__init__(id, *args, **kwargs)
        locations = {
            'W': {'loc': 'left', 'align': ['right', 'center']},
            'N': {'loc': 'top', 'align': ['center', 'bottom']},
            'E': {'loc': 'right', 'align': ['left', 'center']},
            'S': {'loc': 'bottom', 'align': ['center', 'top']}
        }
        self.segments.append(schemdraw.SegmentCircle([0, 0], 0.12, fill='black'))
        self.id_loc = {}
        if isinstance(id_loc, str):
            self.id_loc.update(locations.get(id_loc, {}))
        else:
            self.id_loc.update(id_loc)
        if show:
            self.show()

    def show(self):
        self.segments.append(schemdraw.SegmentCircle([0, 0], 0.12, fill='black'))
        self.bbox = self.get_bbox(includetext=False)
        self.add_label(f'{self.node_id}', **self.id_loc)

    @property
    def name(self) -> str:
        return f'Node {self.node_id}'

class Ground(Node):
    def __init__(self, id: str = '0', *args, **kwargs):
        super().__init__(id, *args, **kwargs)
        gndgap = 0.12
        gnd_lead = 0.4
        resheight = schemdraw.elements.twoterm.resheight
        gap = schemdraw.elements.twoterm.gap
        self.segments.append(schemdraw.Segment(
            [(0, 0), (0, -gnd_lead), (-resheight, -gnd_lead),
             (resheight, -gnd_lead), gap, (-resheight*.7, -gndgap-gnd_lead),
             (resheight*.7, -gndgap-gnd_lead), gap,
             (-resheight*.2, -gndgap*2-gnd_lead),
             (resheight*.2, -gndgap*2-gnd_lead)]))

    @property
    def name(self) -> str:
        return 'Ground'

class VoltageLabel(schemdraw.elements.CurrentLabel):
    def __init__(self, at: schemdraw.elements.Element, label: str = '', **kwargs):
        kwargs['top'] = kwargs.get('top', False)
        kwargs['ofst'] = kwargs.get('ofst', 0.6)
        kwargs['color'] = kwargs.get('color', blue)
        if isinstance(at, RealVoltageSource):
            kwargs['d'] = at.resistor_d
            kwargs['reverse'] = not kwargs.get('reverse', False)
        super().__init__(**kwargs)
        self.params['lblofst'] = 0.1
        if isinstance(at, RealVoltageSource):
            self.at(at.center)
        else:
            self.at(at)
        self.label(label, rotate=kwargs.get('rotate', False))

class CurrentLabel(schemdraw.elements.CurrentLabelInline):
    def __init__(self, at: schemdraw.elements.Element, label: str = '', **kwargs):
        kwargs['color'] = kwargs.get('color', red)
        totlen = at._userparams.get('l', at._userparams.get('unit', 3))
        if isinstance(at, RealVoltageSource):
            kwargs['d'] = at.resistor_d
            totlen = at.resistor_length
        kwargs['ofst'] = totlen/4-0.15+kwargs.get('ofst', 0)
        start = kwargs.get('start', True)
        reverse = kwargs.get('reverse', False)
        if not start and reverse:
            reverse = not reverse
        kwargs.update({'start' : start, 'reverse' : reverse})
        super().__init__(**kwargs)
        if isinstance(at, RealVoltageSource):
            self.at(at.center)
        else:
            self.at(at)
        self.label(label)

    @property
    def name(self) -> str:
        return ''


from dataclasses import dataclass
import schemdraw
from . import Elements as elm
from .ElementTranslators import element_translator, round_node, get_nodes, translator_available
from .Display import red, blue, print_voltage, print_current
from ..Network.network import Network, Branch
from ..Network.solution import NetworkSolver

class UnknownElement(Exception): pass
class MultipleGroundNodes(Exception): pass

def get_node_direction(node1: schemdraw.util.Point, node2: schemdraw.util.Point) -> tuple[int, int]:
    delta = node2 - node1
    delta_x = +1 if delta.x >= 0 else -1
    delta_y = +1 if delta.y >= 0 else -1
    return delta_x, delta_y

@dataclass(frozen=True)
class NetworkDiagramParser:
    drawing: elm.Schematic

    @property
    def all_elements(self) -> list[schemdraw.elements.Element]:
        return self.drawing.elements

    @property
    def circuit_elements(self) -> list[schemdraw.elements.Element]:
        return [e for e in self.all_elements if translator_available(e)]

    @property
    def line_elements(self) -> list[elm.Line]:
        return [e for e in self.all_elements if type(e) is elm.Line]
    
    @property
    def node_elements(self) -> list[elm.Node]:
        return [e for e in self.all_elements if isinstance(e, elm.Node)]

    @property
    def all_nodes(self) -> set[schemdraw.util.Point]:
        nodes = {round_node(e.absanchors['start']) for e in self.circuit_elements}
        nodes = nodes.union({round_node(e.absanchors['end']) for e in self.circuit_elements})
        nodes = nodes.union({round_node(e.absanchors['start']) for e in self.line_elements})
        nodes = nodes.union({round_node(e.absanchors['end']) for e in self.line_elements})
        return nodes

    @property
    def unique_nodes(self) -> set[schemdraw.util.Point]:
        nodes = self.all_nodes
        for node in self.all_nodes:
            if node in nodes:
                for n in self._get_equal_electrical_potential_nodes(node).intersection(nodes):
                    nodes.remove(n)
                nodes.add(node)
        return nodes

    @property
    def node_label_mapping(self) -> dict[schemdraw.util.Point, str]:
        node_labels = {self.unique_node_mapping[get_nodes(e)[0]] : e.node_id for e in self.node_elements}
        node_index = len(node_labels)+1
        unlabeled_nodes = [p for p in self.unique_nodes if p not in node_labels.keys()]
        for p in unlabeled_nodes:
            while str(node_index) in node_labels.values():
                node_index += 1
            node_labels.update({p : str(node_index)})
        return node_labels

    @property
    def unique_node_mapping(self) -> dict[schemdraw.util.Point, schemdraw.util.Point]:
        node_mapping = {}
        for n in self.all_nodes:
            identical_nodes = self._get_equal_electrical_potential_nodes(n)
            identical_nodes.remove(n)
            current_unique_node = self.unique_nodes.intersection(identical_nodes)
            if len(current_unique_node) > 0:
                node_mapping.update({n: current_unique_node.pop()})
            else:
                node_mapping.update({n: n})
        return node_mapping

    @property
    def ground(self) -> schemdraw.util.Point:
        ground_nodes = [n for n in self.node_elements if type(n) == elm.Ground]
        if len(ground_nodes) > 1:
            raise MultipleGroundNodes
        if len(ground_nodes) == 0:
            return list(self.unique_nodes)[0]
        else:
            return get_nodes(ground_nodes[0])[0]

    @property
    def network(self) -> Network:
        translator = lambda e : element_translator[type(e)](e, self._get_node_index)[0]
        return Network(
            branches=[translator(e) for e in self.all_elements if translator_available(e)],
            zero_node_label=self._get_node_index(self.ground)
            )

    def _get_equal_electrical_potential_nodes(self, node: schemdraw.util.Point) -> set[schemdraw.util.Point]:
        equal_electrical_potential_nodes = set([node])
        old_length = 0
        while len(equal_electrical_potential_nodes) > old_length:
            old_length = len(equal_electrical_potential_nodes)
            for line in self.line_elements:
                n1, n2 = get_nodes(line)
                if n1 in equal_electrical_potential_nodes:
                    equal_electrical_potential_nodes.add(n2)
                elif n2 in equal_electrical_potential_nodes:
                    equal_electrical_potential_nodes.add(n1)
        return equal_electrical_potential_nodes

    def _get_node_index(self, node: schemdraw.util.Point) -> str:
        return self.node_label_mapping[self.unique_node_mapping[node]]

    def get_element(self, name: str) -> schemdraw.elements.Element:
        elements = [e for e in self.circuit_elements if e.name == name]
        if len(elements) == 0:
            raise UnknownElement
        else:
            return elements[0]

    def get_branch(self, id: str) -> Branch:
        element = self.get_element(id)
        return element_translator[type(element)](element, self._get_node_index)[0]

class SchemdrawSolution:

    def __init__(self, schemdraw_network: NetworkDiagramParser, solver: NetworkSolver):
        self.schemdraw_network = schemdraw_network
        self.network_solution = solver(self.schemdraw_network.network)

    def draw_voltage(self, name: str, reverse: bool = False, precision: int = 3, top: bool = False) -> elm.VoltageLabel:
        element = self.schemdraw_network.get_element(name)
        branch = self.schemdraw_network.get_branch(name)
        V_branch = self.network_solution.get_voltage(branch)
        if reverse:
            V_branch *= -1
        # adjust counting arrow system of voltage sources for display
        if type(element) is elm.VoltageSource or type(element) is elm.RealVoltageSource:
            reverse = not reverse
        # adjust missing direction information of CurrentLabel() method
        n1, n2 = get_nodes(element)
        dx, dy = get_node_direction(n1, n2)
        if dx < 0 or dy < 0:
            reverse = not reverse
        return elm.VoltageLabel(element, label=f'{print_voltage(V_branch, precision=precision)}V', reverse=reverse, color=blue, top=top)

    def draw_current(self, element_name: str, reverse: bool = False, start: bool = True, ofst: float = 0, precision=3) -> elm.CurrentLabel:
        element = self.schemdraw_network.get_element(element_name)
        branch = self.schemdraw_network.get_branch(element_name)
        I_branch = self.network_solution.get_current(branch)
        if reverse:
            I_branch *= -1
        # adjust counting arrow system of voltage sources for display
        # if type(element) is elm.VoltageSource or type(element) is elm.RealVoltageSource:
        #     start = False
        # if start is False:
        #     reverse = not reverse
        return elm.CurrentLabel(element, label=f'{print_current(I_branch, precision=precision)}A', reverse=reverse, start=start, color=red, ofst=ofst)



from .network import Network, ideal_voltage_sources
from .elements import Element
from dataclasses import dataclass
import numpy as np

class AmbiguousElectricalPotential(Exception): pass

@dataclass(frozen=True)
class SuperNode:
    reference_node: str
    active_node: str
    voltage: complex
    voltage_source: Element

class SuperNodes:
    def __init__(self, network: Network) -> None:
        self._super_nodes : list[SuperNode] = []
        self.voltage_sources: list[Element] = []
        for voltage_source in ideal_voltage_sources(network):
            node, other_node = voltage_source.node1, voltage_source.node2
            if network.is_zero_node(node) or node in self.active_nodes:
                node, other_node = other_node, node
            if network.is_zero_node(node) or node in self.active_nodes:
                raise AmbiguousElectricalPotential
            self.voltage_sources.append(voltage_source.element)
            self._super_nodes.append(
                SuperNode(
                    reference_node=other_node,
                    active_node=node,
                    voltage=voltage_source.element.U if node == voltage_source.node1 else -voltage_source.element.U,
                    voltage_source=voltage_source.element
                )
            )

    @property
    def active_nodes(self) -> list[str]:
        return [sn.active_node for sn in self._super_nodes]

    @property
    def reference_nodes(self) -> list[str]:
        return [sn.reference_node for sn in self._super_nodes]

    @property
    def voltages(self) -> list[complex]:
        return [sn.voltage for sn in self._super_nodes]

    def is_active(self, node: str) -> bool:
        return node in self.active_nodes
    
    def is_reference(self, node: str) -> bool:
        return node in self.reference_nodes

    def belong_to_same(self, active_node: str, reference_node: str) -> bool:
        if self.is_active(active_node):
            if self.is_reference(reference_node):
                return self.get_reference_node(active_node) == reference_node
        return False

    def get_active_node(self, reference_node: str) -> str:
        return self.active_nodes[self.reference_nodes.index(reference_node)]

    def get_reference_node(self, active_node: str) -> str:
        return self.reference_nodes[self.active_nodes.index(active_node)]

    def get_voltage(self, active_node: str) -> complex:
        return self.voltages[self.active_nodes.index(active_node)]

    def sign(self, active_node: str) -> int:
        return np.sign(self.voltages[self.active_nodes.index(active_node)])

    def voltage_to_next_reference(self, active_node: str) -> complex:
        V = 0+0j
        while self.is_active(active_node):
            V += self.voltages[self.active_nodes.index(active_node)]
            active_node = self.get_reference_node(active_node)
        return V

    def next_reference(self, active_node: str) -> str:
        while self.is_active(active_node):
            active_node = self.get_reference_node(active_node)
        return active_node


from typing import Any, Callable
from .network import Network, Branch
from .elements import Element
from . import elements as elm

branch_types : dict[str, Callable[..., Element]] = {
    "resistor" : elm.resistor,
    "real_current_source" : elm.real_current_source,
    "current_source" : elm.current_source,
    "real_voltage_source" : elm.real_voltage_source,
    "voltage_source" : elm.voltage_source,
}

def load_network(network_dict: list[dict[str, Any]]) -> Network:
    branches = []
    for branch in network_dict:
        n1 = branch.pop('N1')
        n2 = branch.pop('N2')
        element_factory = branch_types[branch.pop('type')]
        element = element_factory(**branch)
        branches.append(Branch(n1, n2, element))
    return Network(branches)

def load_network_from_json(filename: str) -> Network:
    import json
    with open(filename, 'r') as json_file:
        network_dict = json.load(json_file)
    return load_network(network_dict)


from typing import Protocol
from dataclasses import dataclass, field
import numpy as np

class Element(Protocol):
    @property
    def Z(self) -> complex:
        """Impedance of Branch"""
    @property
    def Y(self) -> complex:
        """Admittance of Branch"""
    @property
    def I(self) -> complex:
        """Current of Branch"""
    @property
    def U(self) -> complex:
        """Voltage of Branch"""
    @property
    def active(self) -> bool:
        """Whether or not the branch is active"""

@dataclass(frozen=True)
class Impedeance:
    Z : complex
    I : complex = field(default=np.nan, init=False)
    U : complex = field(default=np.nan, init=False)
    active: bool = field(default=False, init=False)
    @property
    def Y(self) -> complex:
        try:
            return 1/self.Z
        except ZeroDivisionError:
            return np.inf

@dataclass(frozen=True)
class RealCurrentSource:
    Z : complex
    I : complex
    active: bool = field(default=True, init=False)
    @property
    def Y(self) -> complex:
        try:
            return 1/self.Z
        except ZeroDivisionError:
            return np.inf
    @property
    def U(self) -> complex:
        try:
            return self.I/self.Y
        except ZeroDivisionError:
            return np.nan

@dataclass(frozen=True)
class CurrentSource:
    Z : complex = field(default=np.inf, init=False)
    Y : complex = field(default=0, init=False)
    I : complex
    U : complex = field(default=np.nan, init=False)
    active: bool = field(default=True, init=False)

@dataclass(frozen=True)
class RealVoltageSource:
    Z : complex
    U : complex
    active: bool = field(default=True, init=False)
    @property
    def Y(self) -> complex:
        try:
            return 1/self.Z
        except ZeroDivisionError:
            return np.inf
    @property
    def I(self) -> complex:
        try:
            return self.U/self.Z
        except ZeroDivisionError:
            return np.nan

@dataclass(frozen=True)
class VoltageSource:
    Z : complex = field(default=0, init=False)
    Y : complex = field(default=np.inf, init=False)
    U : complex
    I : complex = field(default=np.nan, init=False)
    active: bool = field(default=True, init=False)

def resistor(R : float, **_) -> Element:
    return Impedeance(Z=R)

def conductor(G : float, **_) -> Element:
    try:
        return Impedeance(Z=1/G)
    except ZeroDivisionError:
        return Impedeance(Z=np.inf)

def real_current_source(I : float, R : float, **_) -> Element:
    return RealCurrentSource(I=I, Z=R)

def current_source(I : float, **_) -> Element:
    return CurrentSource(I=I)

def real_voltage_source(U : float, R : float, **_) -> Element:
    return RealVoltageSource(U=U, Z=R)

def voltage_source(U : float, **_) -> Element:
    return VoltageSource(U=U)

def is_current_source(element: Element) -> bool:
    return element.active and np.isfinite(element.I)

def is_ideal_voltage_source(element: Element) -> bool:
    return element.active and element.Z==0 and np.isfinite(element.U)

def is_ideal_current_source(element: Element) -> bool:
    return element.active and element.Y==0 and np.isfinite(element.I)


from .network import Network, Branch
from typing import Callable, Protocol

class NetworkSolution(Protocol):
    def get_voltage(self, branch: Branch) -> complex: pass

    def get_current(self, branch: Branch) -> complex: pass

NetworkSolver = Callable[[Network], NetworkSolution]